@@ -1,94 +1,5 @@
 open Import
 
-module Sanitizer : sig
-  [@@@ocaml.warning "-32"]
-
-  module Command : sig
-    type t =
-      { output : string
-      ; build_path_prefix_map : string
-      ; script : Path.t
-      }
-  end
-
-  val impl_sanitizer : (Command.t -> string) -> in_channel -> out_channel -> unit
-
-  val run_sanitizer
-    :  ?temp_dir:Path.t
-    -> prog:Path.t
-    -> argv:string list
-    -> Command.t list
-    -> string list Fiber.t
-end = struct
-  module Command = struct
-    type t =
-      { output : string
-      ; build_path_prefix_map : string
-      ; script : Path.t
-      }
-
-    let of_sexp script (csexp : Sexp.t) : t =
-      match csexp with
-      | List [ Atom build_path_prefix_map; Atom output ] ->
-        { build_path_prefix_map; output; script }
-      | _ -> Code_error.raise "Command.of_csexp: invalid csexp" []
-    ;;
-
-    let to_sexp { output; build_path_prefix_map; script } : Sexp.t =
-      List
-        [ Atom build_path_prefix_map
-        ; Atom output
-        ; Atom (Path.to_absolute_filename script)
-        ]
-    ;;
-  end
-
-  let run_sanitizer ?temp_dir ~prog ~argv commands =
-    let temp_dir =
-      match temp_dir with
-      | Some d -> d
-      | None -> Temp.create Dir ~prefix:"sanitizer" ~suffix:"unspecified"
-    in
-    let fname = Path.relative temp_dir in
-    let stdout_path = fname "sanitizer.stdout" in
-    let stdout_to = Process.Io.file stdout_path Process.Io.Out in
-    let stdin_from =
-      let path = fname "sanitizer.stdin" in
-      let csexp = List.map commands ~f:Command.to_sexp in
-      Io.with_file_out ~binary:true path ~f:(fun oc ->
-        List.iter csexp ~f:(Csexp.to_channel oc));
-      Process.Io.file path Process.Io.In
-    in
-    let open Fiber.O in
-    let+ () = Process.run ~display:Quiet ~stdin_from ~stdout_to Strict prog argv in
-    Io.with_file_in stdout_path ~f:(fun ic ->
-      let rec loop acc =
-        match Csexp.input_opt ic with
-        | Ok None -> List.rev acc
-        | Ok (Some (Sexp.Atom s)) -> loop (s :: acc)
-        | Error error -> Code_error.raise "invalid csexp" [ "error", String error ]
-        | Ok _ -> Code_error.raise "unexpected output" []
-      in
-      loop [])
-  ;;
-
-  let impl_sanitizer f in_ out =
-    set_binary_mode_in in_ true;
-    set_binary_mode_out out true;
-    let rec loop () =
-      match Csexp.input_opt in_ with
-      | Error error -> Code_error.raise "unable to parse csexp" [ "error", String error ]
-      | Ok None -> ()
-      | Ok (Some sexp) ->
-        let command = Command.of_sexp (assert false) sexp in
-        Csexp.to_channel out (Atom (f command));
-        flush out;
-        loop ()
-    in
-    loop ()
-  ;;
-end
-
 (* Translate a path for [sh]. On Windows, [sh] will come from Cygwin so if we
    are a real windows program we need to pass the path through [cygpath] *)
 let translate_path_for_sh =
@@ -96,18 +7,18 @@ let translate_path_for_sh =
   then fun fn -> Fiber.return (Path.to_absolute_filename fn)
   else
     fun fn ->
-    let cygpath =
-      let path = Env_path.path Env.initial in
-      Bin.which ~path "cygpath"
-    in
-    match cygpath with
-    | None -> User_error.raise [ Pp.text "Unable to find cygpath in PATH" ]
-    | Some cygpath ->
-      Process.run_capture_line
-        ~display:Quiet
-        Strict
-        cygpath
-        [ Path.to_absolute_filename fn ]
+      let cygpath =
+        let path = Env_path.path Env.initial in
+        Bin.which ~path "cygpath"
+      in
+      match cygpath with
+      | None -> User_error.raise [ Pp.text "Unable to find cygpath in PATH" ]
+      | Some cygpath ->
+        Process.run_capture_line
+          ~display:Quiet
+          Strict
+          cygpath
+          [ Path.to_absolute_filename fn ]
 ;;
 
 (* Quote a filename for sh, independently of whether we are on Windows or Unix.
@@ -196,10 +107,23 @@ type metadata_entry =
   ; build_path_prefix_map : string
   }
 
+let dyn_of_metadata_entry { exit_code; build_path_prefix_map } =
+  let open Dyn in
+  record
+    [ "exit_code", int exit_code; "build_path_prefix_map", string build_path_prefix_map ]
+;;
+
 type metadata_result =
   | Present of metadata_entry
   | Missing_unreachable
 
+let dyn_of_metadata_result =
+  let open Dyn in
+  function
+  | Missing_unreachable -> variant "Missing_unreachable" []
+  | Present p -> variant "Present" [ dyn_of_metadata_entry p ]
+;;
+
 type full_block_result = block_result * metadata_result
 
 type sh_script =
@@ -214,7 +138,7 @@ let read_exit_codes_and_prefix_maps file =
     | None -> ""
     | Some file ->
       (try Io.read_file ~binary:true file with
-       | Sys_error _ ->
+       | Unix.Unix_error _ | Sys_error _ ->
          (* a script where the first command immediately exits might not produce
             the metadata file *)
          "")
@@ -286,9 +210,22 @@ let rewrite_paths build_path_prefix_map ~parent_script ~command_script s =
     |> Re.replace_string error_msg ~by:""
 ;;
 
-let sanitize ~parent_script cram_to_output
-  : (block_result * metadata_result * string) Cram_lexer.block list
-  =
+type command_out =
+  { command : string list
+  ; metadata : metadata_result
+  ; output : string
+  }
+
+let dyn_of_command_out { command; metadata; output } =
+  let open Dyn in
+  record
+    [ "command", (list string) command
+    ; "metadata", dyn_of_metadata_result metadata
+    ; "output", string output
+    ]
+;;
+
+let sanitize ~parent_script cram_to_output : command_out Cram_lexer.block list =
   List.map cram_to_output ~f:(fun (t : (block_result * _) Cram_lexer.block) ->
     match t with
     | Cram_lexer.Comment t -> Cram_lexer.Comment t
@@ -304,7 +241,7 @@ let sanitize ~parent_script cram_to_output
                ~command_script:block_result.script
                build_path_prefix_map
       in
-      Command (block_result, metadata, output))
+      Command { command = block_result.command; metadata; output })
 ;;
 
 (* Compose user written cram stanzas to output *)
@@ -321,7 +258,7 @@ let compose_cram_output (cram_to_output : _ Cram_lexer.block list) =
   List.iter cram_to_output ~f:(fun block ->
     match (block : _ Cram_lexer.block) with
     | Comment lines -> List.iter lines ~f:add_line
-    | Command ({ command; output_file = _; script = _ }, metadata, output) ->
+    | Command { command; metadata; output } ->
       List.iteri command ~f:(fun i line ->
         let line = sprintf "%c %s" (if i = 0 then '$' else '>') line in
         add_line_prefixed_with_two_space line);
@@ -333,9 +270,29 @@ let compose_cram_output (cram_to_output : _ Cram_lexer.block list) =
   Buffer.contents buf
 ;;
 
+(* Compose user written cram stanzas to output *)
+let cram_commmands commands =
+  let buf = Buffer.create 256 in
+  let add_line line =
+    Buffer.add_string buf line;
+    Buffer.add_char buf '\n'
+  in
+  let add_line_prefixed_with_two_space line =
+    Buffer.add_string buf "  ";
+    add_line line
+  in
+  List.iter commands ~f:(fun command ->
+    List.iteri command ~f:(fun i line ->
+      let line = sprintf "%c %s" (if i = 0 then '$' else '>') line in
+      add_line_prefixed_with_two_space line));
+  Buffer.contents buf
+;;
+
 let create_sh_script cram_stanzas ~temp_dir : sh_script Fiber.t =
   let script = Path.relative temp_dir "main.sh" in
   let oc = Io.open_out ~binary:true script in
+  Fiber.finalize ~finally:(fun () -> Fiber.return @@ close_out oc)
+  @@ fun () ->
   let file name = Path.relative temp_dir name in
   let open Fiber.O in
   let sh_path path =
@@ -381,7 +338,6 @@ let create_sh_script cram_stanzas ~temp_dir : sh_script Fiber.t =
   in
   fprln oc "trap 'exit 0' EXIT";
   let+ cram_to_output = Fiber.sequential_map ~f:loop cram_stanzas in
-  close_out oc;
   let command_count = !i in
   let metadata_file = Option.some_if (command_count > 0) metadata_file in
   { script; cram_to_output; metadata_file }
@@ -389,7 +345,21 @@ let create_sh_script cram_stanzas ~temp_dir : sh_script Fiber.t =
 
 let _display_with_bars s = List.iter (String.split_lines s) ~f:(Printf.eprintf "| %s\n")
 
-let run ~env ~script lexbuf : string Fiber.t =
+let make_run_env env ~temp_dir ~cwd =
+  let env = Env.add env ~var:"LC_ALL" ~value:"C" in
+  let temp_dir = Path.relative temp_dir "tmp" in
+  let env =
+    Dune_util.Build_path_prefix_map.extend_build_path_prefix_map
+      env
+      `New_rules_have_precedence
+      [ Some { source = Path.to_absolute_filename cwd; target = "$TESTCASE_ROOT" }
+      ; Some { source = Path.to_absolute_filename temp_dir; target = "$TMPDIR" }
+      ]
+  in
+  Env.add env ~var:Env.Var.temp_dir ~value:(Path.to_absolute_filename temp_dir)
+;;
+
+let make_temp_dir ~script =
   let temp_dir =
     let suffix =
       let basename = Path.basename script in
@@ -402,79 +372,230 @@ let run ~env ~script lexbuf : string Fiber.t =
     in
     Temp.create Dir ~prefix:"dune_cram" ~suffix
   in
-  let cram_stanzas = cram_stanzas lexbuf in
+  Path.mkdir_p temp_dir;
+  temp_dir
+;;
+
+let run_cram_test env ~src ~script ~cram_stanzas ~temp_dir ~cwd ~timeout =
   let open Fiber.O in
   let* sh_script = create_sh_script cram_stanzas ~temp_dir in
-  let cwd = Path.parent_exn script in
-  let env =
-    let env = Env.add env ~var:"LC_ALL" ~value:"C" in
-    let temp_dir = Path.relative temp_dir "tmp" in
-    let env =
-      Dune_util.Build_path_prefix_map.extend_build_path_prefix_map
-        env
-        `New_rules_have_precedence
-        [ Some { source = Path.to_absolute_filename cwd; target = "$TESTCASE_ROOT" }
-        ; Some { source = Path.to_absolute_filename temp_dir; target = "$TMPDIR" }
-        ]
-    in
-    Path.mkdir_p temp_dir;
-    Env.add env ~var:Env.Var.temp_dir ~value:(Path.to_absolute_filename temp_dir)
-  in
+  let env = make_run_env env ~temp_dir ~cwd in
   let open Fiber.O in
-  let+ () =
-    let sh =
-      let path = Env_path.path Env.initial in
-      Option.value_exn (Bin.which ~path "sh")
-    in
-    let metadata =
-      let name =
-        let base = Path.basename sh_script.script in
-        match String.equal base "run.t" with
-        | false -> base
-        | true -> sprintf "%s/%s" (Path.basename (Path.parent_exn sh_script.script)) base
-      in
-      Process.create_metadata ~name ~categories:[ "cram" ] ()
-    in
-    Process.run
-      ~display:Quiet
-      ~metadata
-      ~dir:cwd
-      ~env
-      Strict
-      sh
-      [ Path.to_string sh_script.script ]
+  let sh =
+    let path = Env_path.path Env.initial in
+    match Bin.which ~path "sh" with
+    | Some sh -> sh
+    | None ->
+      User_error.raise [ Pp.text "CRAM test aborted, \"sh\" can not be found in PATH" ]
   in
-  let raw = read_and_attach_exit_codes sh_script in
-  let sanitized = sanitize ~parent_script:sh_script.script raw in
-  compose_cram_output sanitized
+  let metadata =
+    let name =
+      let base = Path.basename sh_script.script in
+      match String.equal base "run.t" with
+      | false -> base
+      | true -> sprintf "%s/%s" (Path.basename (Path.parent_exn sh_script.script)) base
+    in
+    Process.create_metadata ~name ~categories:[ "cram" ] ()
+  in
+  Process.run
+    ~display:Quiet
+    ~metadata
+    ~dir:cwd
+    ~env
+    (Timeout { timeout_seconds = Option.map ~f:snd timeout; failure_mode = Strict })
+    sh
+    [ Path.to_string sh_script.script ]
+  >>| function
+  | Ok () -> read_and_attach_exit_codes sh_script |> sanitize ~parent_script:script
+  | Error `Timed_out ->
+    let timeout_loc, timeout = Option.value_exn timeout in
+    User_error.raise
+      ~loc:(Loc.in_file (Path.drop_optional_build_context_maybe_sandboxed src))
+      [ Pp.concat
+          [ Pp.paragraphf
+              "Cram test timed out. A time limit of %.2fs has been set in "
+              timeout
+          ; Pp.tag User_message.Style.Loc @@ Loc.pp_file_colon_line timeout_loc
+          ; Pp.verbatim "."
+          ]
+      ]
 ;;
 
-let run ~env ~script = run_expect_test script ~f:(fun lexbuf -> run ~env ~script lexbuf)
+let run_produce_correction ~src ~env ~script ~timeout lexbuf =
+  let temp_dir = make_temp_dir ~script in
+  let cram_stanzas = cram_stanzas lexbuf in
+  let cwd = Path.parent_exn script in
+  let env = make_run_env env ~temp_dir ~cwd in
+  let open Fiber.O in
+  run_cram_test env ~src ~script ~cram_stanzas ~temp_dir ~cwd ~timeout
+  >>| compose_cram_output
+;;
 
-module Spec = struct
-  type ('path, _) t = 'path
+module Script = Persistent.Make (struct
+    type nonrec t = command_out list
 
-  let name = "cram"
-  let version = 1
-  let bimap path f _ = f path
-  let is_useful_to ~memoize:_ = true
+    let name = "CRAM-RESULT"
+    let version = 1
+    let to_dyn = Dyn.list dyn_of_command_out
+    let test_example () = []
+  end)
 
-  let encode script path _ : Dune_lang.t =
-    List [ Dune_lang.atom_or_quoted_string "cram"; path script ]
-  ;;
+let run_and_produce_output ~src ~env ~dir:cwd ~script ~dst ~timeout =
+  let script_contents = Io.read_file ~binary:false script in
+  let lexbuf = Lexbuf.from_string script_contents ~fname:(Path.to_string script) in
+  let temp_dir = make_temp_dir ~script in
+  let cram_stanzas = cram_stanzas lexbuf in
+  Path.unlink_exn script;
+  let env = make_run_env env ~temp_dir ~cwd in
+  let open Fiber.O in
+  run_cram_test env ~src ~script ~cram_stanzas ~temp_dir ~cwd ~timeout
+  >>| List.filter_map ~f:(function
+    | Cram_lexer.Command c -> Some c
+    | Comment _ -> None)
+  >>| Script.dump (Path.build dst)
+;;
 
-  let action script ~ectx:_ ~(eenv : Action.Ext.env) = run ~env:eenv.env ~script
+module Run = struct
+  module Spec = struct
+    type ('path, 'target) t =
+      { src : Path.t
+      ; dir : 'path
+      ; script : 'path
+      ; output : 'target
+      ; timeout : (Loc.t * float) option
+      }
+
+    let name = "cram-run"
+    let version = 2
+
+    let bimap ({ src = _; dir; script; output; timeout } as t) f g =
+      { t with dir = f dir; script = f script; output = g output; timeout }
+    ;;
+
+    let is_useful_to ~memoize:_ = true
+
+    let encode { src = _; dir; script; output; timeout } path target : Sexp.t =
+      List
+        [ path dir
+        ; path script
+        ; target output
+        ; Dune_sexp.Encoder.(option float (Option.map ~f:snd timeout))
+          |> Dune_sexp.to_sexp
+        ]
+    ;;
+
+    let action { src; dir; script; output; timeout } ~ectx:_ ~(eenv : Action.env) =
+      run_and_produce_output ~src ~env:eenv.env ~dir ~script ~dst:output ~timeout
+    ;;
+  end
+
+  include Action_ext.Make (Spec)
 end
 
-let action script =
-  let module M = struct
-    type path = Path.t
-    type target = Path.Build.t
-
-    module Spec = Spec
-
-    let v = script
-  end
-  in
-  Action.Extension (module M)
+let run ~src ~dir ~script ~output ~timeout =
+  Run.action { src; dir; script; output; timeout }
 ;;
+
+module Make_script = struct
+  module Spec = struct
+    type ('path, 'target) t = 'path * 'target
+
+    let name = "cram-generate"
+    let version = 1
+    let bimap (src, dst) f g = f src, g dst
+    let is_useful_to ~memoize:_ = true
+    let encode (src, dst) path target : Sexp.t = List [ path src; target dst ]
+
+    let action (src, dst) ~ectx:_ ~eenv:_ =
+      let commands =
+        Io.read_file ~binary:false src
+        |> Lexbuf.from_string ~fname:(Path.to_string src)
+        |> cram_stanzas
+        |> List.filter_map ~f:(function
+          | Cram_lexer.Comment _ -> None
+          | Command s -> Some s)
+        |> cram_commmands
+      in
+      Io.write_file ~binary:false (Path.build dst) commands;
+      Fiber.return ()
+    ;;
+  end
+
+  include Action_ext.Make (Spec)
+end
+
+let make_script ~src ~script = Make_script.action (src, script)
+
+module Diff = struct
+  module Spec = struct
+    type ('path, _) t =
+      { script : 'path
+      ; out : 'path
+      }
+
+    let name = "cram-generate"
+    let version = 1
+    let bimap { script; out } f _ = { script = f script; out = f out }
+    let is_useful_to ~memoize:_ = true
+    let encode { script; out } path _ : Sexp.t = List [ path script; path out ]
+
+    let action { script; out } ~ectx:_ ~eenv:_ =
+      let current = Io.read_file ~binary:false script in
+      let combined =
+        let out =
+          match Script.load out with
+          | Some s -> s
+          | None ->
+            User_error.raise
+              [ Pp.textf "%s does not exist or is corrupted" (Path.to_string out) ]
+        in
+        let current_stanzas =
+          Lexbuf.from_string ~fname:(Path.to_string script) current |> cram_stanzas
+        in
+        let rec loop acc current expected =
+          match current with
+          | [] -> acc
+          | Cram_lexer.Comment x :: current ->
+            loop (Cram_lexer.Comment x :: acc) current expected
+          | Command _ :: current ->
+            (match expected with
+             | [] -> acc
+             | out :: expected -> loop (Cram_lexer.Command out :: acc) current expected)
+        in
+        loop [] current_stanzas out |> List.rev
+      in
+      let expected = compose_cram_output combined in
+      let corrected_file = Path.extend_basename script ~suffix:".corrected" in
+      if String.equal current expected
+      then Path.rm_rf corrected_file
+      else Io.write_file ~binary:false corrected_file expected;
+      Fiber.return ()
+    ;;
+  end
+
+  include Action_ext.Make (Spec)
+end
+
+let diff ~src ~output = Diff.action { script = src; out = output }
+
+module Action = struct
+  module Spec = struct
+    type ('path, _) t = 'path
+
+    let name = "cram"
+    let version = 2
+    let bimap path f _ = f path
+    let is_useful_to ~memoize:_ = true
+    let encode script path _ : Sexp.t = List [ path script ]
+
+    let action script ~ectx:_ ~(eenv : Action.env) =
+      run_expect_test
+        script
+        ~f:(run_produce_correction ~src:script ~env:eenv.env ~script ~timeout:None)
+    ;;
+  end
+
+  include Action_ext.Make (Spec)
+end
+
+let action = Action.action
